from abc import ABC, abstractmethod
from collections.abc import Collection
from typing import Literal

from anndata import AnnData
from pandas import DataFrame

from ngio.utils import ZarrGroupHandler


class AbstractTableBackend(ABC):
    """Abstract class for table backends."""

    def __init__(
        self,
        group_handler: ZarrGroupHandler,
        index_key: str | None = None,
        index_type: Literal["int", "str"] = "int",
    ):
        """Initialize the handler.

        Args:
            group_handler (ZarrGroupHandler): An object to handle the Zarr group
                containing the table data.
            index_key (str): The column name to use as the index of the DataFrame.
            index_type (str): The type of the index column in the DataFrame.
        """
        self._group_handler = group_handler
        self._index_key = index_key
        self._index_type = index_type

    @staticmethod
    @abstractmethod
    def backend_name() -> str:
        """The name of the backend."""
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def implements_anndata() -> bool:
        """Whether the handler implements the anndata protocol."""
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def implements_dataframe() -> bool:
        """Whether the handler implements the dataframe protocol."""
        raise NotImplementedError

    @abstractmethod
    def load_columns(self) -> list[str]:
        """List all labels in the group."""
        raise NotImplementedError

    def load_as_anndata(self, columns: Collection[str] | None = None) -> AnnData:
        """Load the metadata in the store."""
        raise NotImplementedError

    def load_as_dataframe(self, columns: Collection[str] | None = None) -> DataFrame:
        """List all labels in the group."""
        raise NotImplementedError

    def write_from_dataframe(
        self, table: DataFrame, metadata: dict | None = None
    ) -> None:
        """Consolidate the metadata in the store."""
        raise NotImplementedError

    def write_from_anndata(self, table: AnnData, metadata: dict | None = None) -> None:
        """Consolidate the metadata in the store."""
        raise NotImplementedError



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal

import anndata as ad
import numpy as np
import pandas as pd
import pandas.api.types as ptypes
import zarr
from anndata import AnnData
from anndata._io.specs import read_elem
from anndata._io.utils import _read_legacy_raw
from anndata._io.zarr import read_dataframe
from anndata.compat import _clean_uns
from anndata.experimental import read_dispatched

from ngio.tables._validators import validate_index_dtype, validate_index_key
from ngio.utils import (
    NgioTableValidationError,
    NgioValueError,
    StoreOrGroup,
    open_group_wrapper,
)

if TYPE_CHECKING:
    from collections.abc import Callable, Collection


def custom_read_zarr(
    store: StoreOrGroup, elem_to_read: Collection[str] | None = None
) -> AnnData:
    """Read from a hierarchical Zarr array store.

    # Implementation originally from https://github.com/scverse/anndata/blob/main/src/anndata/_io/zarr.py
    # Original implementation would not work with remote storages so we had to copy it
    # here and slightly modified it to work with remote storages.

    Args:
        store (StoreOrGroup): A store or group to read the AnnData from.
        elem_to_read (Collection[str] | None): The elements to read from the store.
    """
    group, _ = open_group_wrapper(store=store, mode="r")

    if not isinstance(group.store, zarr.DirectoryStore):
        elem_to_read = ["X", "obs", "var"]

    if elem_to_read is None:
        elem_to_read = [
            "X",
            "obs",
            "var",
            "uns",
            "obsm",
            "varm",
            "obsp",
            "varp",
            "layers",
        ]

    # Read with handling for backwards compat
    def callback(func: Callable, elem_name: str, elem: Any, iospec: Any) -> Any:
        if iospec.encoding_type == "anndata" or elem_name.endswith("/"):
            ad_kwargs = {}
            # Some of these elem fail on https
            # So we only include the ones that are strictly necessary
            # for fractal tables
            # This fails on some https
            # base_elem += list(elem.keys())
            for k in elem_to_read:
                v = elem.get(k)
                if v is not None and not k.startswith("raw."):
                    ad_kwargs[k] = read_dispatched(v, callback)  # type: ignore
            return AnnData(**ad_kwargs)

        elif elem_name.startswith("/raw."):
            return None
        elif elem_name in {"/obs", "/var"}:
            return read_dataframe(elem)
        elif elem_name == "/raw":
            # Backwards compat
            return _read_legacy_raw(group, func(elem), read_dataframe, func)
        return func(elem)

    adata = read_dispatched(group, callback=callback)  # type: ignore

    # Backwards compat (should figure out which version)
    if "raw.X" in group:
        raw = AnnData(**_read_legacy_raw(group, adata.raw, read_dataframe, read_elem))  # type: ignore
        raw.obs_names = adata.obs_names  # type: ignore
        adata.raw = raw  # type: ignore

    # Backwards compat for <0.7
    if isinstance(group["obs"], zarr.Array):
        _clean_uns(adata)

    if not isinstance(adata, AnnData):
        raise NgioValueError(f"Expected an AnnData object, but got {type(adata)}")
    return adata


def _check_for_mixed_types(series: pd.Series) -> None:
    """Check if the column has mixed types."""
    if series.apply(type).nunique() > 1:  # type: ignore
        raise NgioTableValidationError(
            f"Column {series.name} has mixed types: "
            f"{series.apply(type).unique()}. "  # type: ignore
            "Type of all elements must be the same."
        )


def _check_for_supported_types(series: pd.Series) -> Literal["str", "int", "numeric"]:
    """Check if the column has supported types."""
    if ptypes.is_string_dtype(series):
        return "str"
    if ptypes.is_integer_dtype(series):
        return "int"
    if ptypes.is_numeric_dtype(series):
        return "numeric"
    raise NgioTableValidationError(
        f"Column {series.name} has unsupported type: {series.dtype}."
        " Supported types are string and numerics."
    )


def dataframe_to_anndata(
    dataframe: pd.DataFrame,
    index_key: str | None = None,
    overwrite: bool = False,
) -> ad.AnnData:
    """Convert a table DataFrame to an AnnData object.

    Args:
        dataframe (pd.DataFrame): A pandas DataFrame representing a fractal table.
        index_key (str): The column name to use as the index of the DataFrame.
            Default is None.
        overwrite (bool): Whether to overwrite the index if a different index is found.
            Default is False.
    """
    # Check if the index_key is present in the data frame + optional validations
    dataframe = validate_index_key(dataframe, index_key, overwrite=overwrite)
    dataframe = validate_index_dtype(dataframe, index_type="str")

    str_columns, int_columns, num_columns = [], [], []
    for c_name in dataframe.columns:
        column_df = dataframe[c_name]
        _check_for_mixed_types(column_df)  # Mixed types are not allowed in the table
        c_type = _check_for_supported_types(
            column_df
        )  # Only string and numeric types are allowed

        if c_type == "str":
            str_columns.append(c_name)

        elif c_type == "int":
            int_columns.append(c_name)

        elif c_type == "numeric":
            num_columns.append(c_name)

    # Converting all observations to string
    obs_df = dataframe[str_columns + int_columns]
    obs_df.index = dataframe.index

    x_df = dataframe[num_columns]

    if x_df.dtypes.nunique() > 1:
        x_df = x_df.astype("float64")

    if x_df.empty:
        # If there are no numeric columns, create an empty array
        # to avoid AnnData failing to create the object
        x_df = np.zeros((len(obs_df), 0), dtype="float64")

    return ad.AnnData(X=x_df, obs=obs_df)


def anndata_to_dataframe(
    anndata: ad.AnnData,
    index_key: str | None = "label",
    index_type: str = "int",
    overwrite: bool = False,
) -> pd.DataFrame:
    """Convert a AnnData object representing a fractal table to a pandas DataFrame.

    Args:
        anndata (ad.AnnData): An AnnData object representing a fractal table.
        index_key (str): The column name to use as the index of the DataFrame.
            Default is 'label'.
        index_type (str): The type of the index column in the DataFrame.
            Either 'str' or 'int'. Default is 'int'.
        overwrite (bool): Whether to overwrite the index if a different index is found.
            Default is False.
    """
    dataframe = anndata.to_df()
    dataframe[anndata.obs_keys()] = anndata.obs
    dataframe = validate_index_key(dataframe, index_key, overwrite=overwrite)
    dataframe = validate_index_dtype(dataframe, index_type)
    return dataframe



"""Ngio Tables backend implementations."""

from ngio.tables.backends._table_backends import (
    ImplementedTableBackends,
    TableBackendProtocol,
)

__all__ = ["ImplementedTableBackends", "TableBackendProtocol"]



from pathlib import Path

import pytest

from ngio import create_empty_plate, open_omezarr_plate
from ngio.utils import NgioValueError


def test_open_real_omezarr_plate(cardiomyocyte_tiny_path: Path):
    cardiomyocyte_tiny_path = cardiomyocyte_tiny_path
    omezarr_plate = open_omezarr_plate(cardiomyocyte_tiny_path)

    assert isinstance(omezarr_plate.__repr__(), str)
    assert omezarr_plate.columns == ["03"]
    assert omezarr_plate.rows == ["B"]
    assert omezarr_plate.acquisitions_ids == [0]
    assert omezarr_plate.acquisitions_names == [
        "20200812-CardiomyocyteDifferentiation14-Cycle1"
    ]

    well_path = omezarr_plate.get_well_path("B", "03")
    well_path2 = omezarr_plate.get_well_path("B", 3)
    assert well_path == well_path2
    well = omezarr_plate.get_well("B", "03")
    assert well.paths() == ["0"]

    image_path = omezarr_plate.get_image_path("B", "03", "0")
    assert image_path == "B/03/0"

    images_plate = omezarr_plate.get_images()
    images_well = omezarr_plate.get_well_images("B", "03")
    assert len(images_plate) == 1
    assert len(images_well) == 1


def test_create_and_edit_plate(tmp_path: Path):
    test_plate = create_empty_plate(tmp_path / "test_plate.zarr", name="test_plate")
    assert test_plate.columns == []
    assert test_plate.rows == []
    assert test_plate.acquisitions_ids == []

    test_plate.add_image(row="B", column="03", image_path="0", acquisition_id=0)
    test_plate.add_image(row="B", column="03", image_path="1", acquisition_id=0)

    with pytest.raises(NgioValueError):
        test_plate.add_image(row="B", column="03", image_path="1", acquisition_id=1)

    test_plate.atomic_add_image(row="C", column="02", image_path="1", acquisition_id=1)

    assert test_plate.columns == ["02", "03"]
    assert test_plate.rows == ["B", "C"]
    assert test_plate.acquisitions_ids == [0, 1]

    assert len(test_plate.wells_paths) == 2

    test_plate.remove_image(row="C", column="02", image_path="1")
    assert len(test_plate.wells_paths) == 1



"""A module for handling the Plate Collection in an OME-Zarr file."""

from ngio.images import OmeZarrContainer
from ngio.ome_zarr_meta import (
    ImageInWellPath,
    NgioPlateMeta,
    NgioWellMeta,
    find_plate_meta_handler,
    find_well_meta_handler,
    get_plate_meta_handler,
    get_well_meta_handler,
)
from ngio.utils import (
    AccessModeLiteral,
    StoreOrGroup,
    ZarrGroupHandler,
)


# Mock lock class that does nothing
class MockLock:
    """A mock lock class that does nothing."""

    def __enter__(self):
        """Enter the lock."""
        pass

    def __exit__(self, exc_type, exc_value, traceback):
        """Exit the lock."""
        pass


class OmeZarrWell:
    """A class to handle the Well Collection in an OME-Zarr file."""

    def __init__(self, group_handler: ZarrGroupHandler) -> None:
        """Initialize the LabelGroupHandler.

        Args:
            group_handler: The Zarr group handler that contains the Well.
        """
        self._group_handler = group_handler
        self._meta_handler = find_well_meta_handler(group_handler)

    @property
    def meta_handler(self):
        """Return the metadata handler."""
        return self._meta_handler

    @property
    def meta(self):
        """Return the metadata."""
        return self._meta_handler.meta

    def paths(self, acquisition: int | None = None) -> list[str]:
        """Return the images paths in the well.

        If acquisition is None, return all images paths in the well.
        Else, return the images paths in the well for the given acquisition.

        Args:
            acquisition (int | None): The acquisition id to filter the images.
        """
        return self.meta.paths(acquisition)


class OmeZarrPlate:
    """A class to handle the Plate Collection in an OME-Zarr file."""

    def __init__(self, group_handler: ZarrGroupHandler) -> None:
        """Initialize the LabelGroupHandler.

        Args:
            group_handler: The Zarr group handler that contains the Plate.
        """
        self._group_handler = group_handler
        self._meta_handler = find_plate_meta_handler(group_handler)

    def __repr__(self) -> str:
        """Return a string representation of the plate."""
        return f"Plate([rows x columns] ({len(self.rows)} x {len(self.columns)})"

    @property
    def meta_handler(self):
        """Return the metadata handler."""
        return self._meta_handler

    @property
    def meta(self):
        """Return the metadata."""
        return self._meta_handler.meta

    @property
    def columns(self) -> list[str]:
        """Return the number of columns in the plate."""
        return self.meta.columns

    @property
    def rows(self) -> list[str]:
        """Return the number of rows in the plate."""
        return self.meta.rows

    @property
    def acquisitions_names(self) -> list[str | None]:
        """Return the acquisitions in the plate."""
        return self.meta.acquisitions_names

    @property
    def acquisitions_ids(self) -> list[int]:
        """Return the acquisitions ids in the plate."""
        return self.meta.acquisitions_ids

    @property
    def wells_paths(self) -> list[str]:
        """Return the wells paths in the plate."""
        return self.meta.wells_paths

    def get_well_path(self, row: str, column: int | str) -> str:
        """Return the well path in the plate."""
        return self.meta.get_well_path(row=row, column=column)

    def get_image_path(self, row: str, column: int | str, path: str) -> str:
        """Return the image path in the plate."""
        well = self.get_well(row, column)
        if path not in well.paths():
            raise ValueError(f"Image {path} does not exist in well {row}{column}")
        return f"{self.get_well_path(row, column)}/{path}"

    def get_well(self, row: str, column: int | str) -> OmeZarrWell:
        """Get a well from the plate.

        Args:
            row (str): The row of the well.
            column (int | str): The column of the well.

        Returns:
            OmeZarrWell: The well.
        """
        well_path = self.meta.get_well_path(row=row, column=column)
        group_handler = self._group_handler.derive_handler(well_path)
        return OmeZarrWell(group_handler)

    def get_wells(self) -> dict[str, OmeZarrWell]:
        """Get all wells in the plate."""
        wells = {}
        for well_path in self.wells_paths:
            group_handler = self._group_handler.derive_handler(well_path)
            well = OmeZarrWell(group_handler)
            wells[well_path] = well
        return wells

    def get_images(self, acquisition: int | None = None) -> list[OmeZarrContainer]:
        """Get all images in the plate.

        Args:
            acquisition: The acquisition id to filter the images.
        """
        images = []
        for well_path, well in self.get_wells().items():
            for img_path in well.paths(acquisition):
                full_path = f"{well_path}/{img_path}"
                img_group_handler = self._group_handler.derive_handler(full_path)
                images.append(OmeZarrContainer(img_group_handler))
        return images

    def get_well_images(
        self, row: str, column: str | int, acquisition: int | None = None
    ) -> list[OmeZarrContainer]:
        """Get all images in a well.

        Args:
            row: The row of the well.
            column: The column of the well.
            acquisition: The acquisition id to filter the images.
        """
        well_path = self.meta.get_well_path(row=row, column=column)
        group_handler = self._group_handler.derive_handler(well_path)
        well = OmeZarrWell(group_handler)

        images = []
        for path in well.paths(acquisition):
            image_path = f"{well_path}/{path}"
            group_handler = self._group_handler.derive_handler(image_path)
            images.append(OmeZarrContainer(group_handler))

        return images

    def _add_image(
        self,
        row: str,
        column: int | str,
        image_path: str,
        acquisition_id: int | None = None,
        acquisition_name: str | None = None,
        atomic: bool = False,
    ) -> StoreOrGroup:
        """Add an image to an ome-zarr plate."""
        if atomic:
            plate_lock = self._group_handler.lock
        else:
            plate_lock = MockLock()

        with plate_lock:
            meta = self.meta
            meta = meta.add_well(row, column, acquisition_id, acquisition_name)
            self.meta_handler.write_meta(meta)
            self.meta_handler._group_handler.clean_cache()

        well_path = self.meta.get_well_path(row=row, column=column)
        group_handler = self._group_handler.derive_handler(well_path)

        if atomic:
            well_lock = group_handler.lock
        else:
            well_lock = MockLock()

        with well_lock:
            attrs = group_handler.load_attrs()
            if len(attrs) == 0:
                # Initialize the well metadata
                # if the group is empty
                well_meta = NgioWellMeta.default_init()
                meta_handler = get_well_meta_handler(group_handler, version="0.4")
            else:
                meta_handler = find_well_meta_handler(group_handler)
                well_meta = meta_handler.meta

            group_handler = self._group_handler.derive_handler(well_path)

            well_meta = well_meta.add_image(path=image_path, acquisition=acquisition_id)
            meta_handler.write_meta(well_meta)
            meta_handler._group_handler.clean_cache()

        return group_handler.get_group(image_path, create_mode=True)

    def atomic_add_image(
        self,
        row: str,
        column: int | str,
        image_path: str,
        acquisition_id: int | None = None,
        acquisition_name: str | None = None,
    ) -> StoreOrGroup:
        """Parallel safe version of add_image."""
        return self._add_image(
            row=row,
            column=column,
            image_path=image_path,
            acquisition_id=acquisition_id,
            acquisition_name=acquisition_name,
            atomic=True,
        )

    def add_image(
        self,
        row: str,
        column: int | str,
        image_path: str,
        acquisition_id: int | None = None,
        acquisition_name: str | None = None,
    ) -> StoreOrGroup:
        """Add an image to an ome-zarr plate."""
        return self._add_image(
            row=row,
            column=column,
            image_path=image_path,
            acquisition_id=acquisition_id,
            acquisition_name=acquisition_name,
            atomic=False,
        )

    def _remove_well(
        self,
        row: str,
        column: int | str,
        atomic: bool = False,
    ):
        """Remove a well from an ome-zarr plate."""
        if atomic:
            plate_lock = self._group_handler.lock
        else:
            plate_lock = MockLock()

        with plate_lock:
            meta = self.meta
            meta = meta.remove_well(row, column)
            self.meta_handler.write_meta(meta)
            self.meta_handler._group_handler.clean_cache()

    def _remove_image(
        self,
        row: str,
        column: int | str,
        image_path: str,
        atomic: bool = False,
    ):
        """Remove an image from an ome-zarr plate."""
        well = self.get_well(row, column)

        if atomic:
            well_lock = well.meta_handler._group_handler.lock
        else:
            well_lock = MockLock()

        with well_lock:
            well_meta = well.meta
            well_meta = well_meta.remove_image(path=image_path)
            well.meta_handler.write_meta(well_meta)
            well.meta_handler._group_handler.clean_cache()
            if len(well_meta.paths()) == 0:
                self._remove_well(row, column, atomic=atomic)

    def atomic_remove_image(
        self,
        row: str,
        column: int | str,
        image_path: str,
    ):
        """Parallel safe version of remove_image."""
        return self._remove_image(
            row=row,
            column=column,
            image_path=image_path,
            atomic=True,
        )

    def remove_image(
        self,
        row: str,
        column: int | str,
        image_path: str,
    ):
        """Remove an image from an ome-zarr plate."""
        return self._remove_image(
            row=row,
            column=column,
            image_path=image_path,
            atomic=False,
        )


def open_omezarr_plate(
    store: StoreOrGroup,
    cache: bool = False,
    mode: AccessModeLiteral = "r+",
    parallel_safe: bool = True,
) -> OmeZarrPlate:
    """Open an OME-Zarr plate.

    Args:
        store (StoreOrGroup): The Zarr store or group that stores the plate.
        cache (bool): Whether to use a cache for the zarr group metadata.
        mode (AccessModeLiteral): The
            access mode for the image. Defaults to "r+".
        parallel_safe (bool): Whether the group handler is parallel safe.
    """
    group_handler = ZarrGroupHandler(
        store=store, cache=cache, mode=mode, parallel_safe=parallel_safe
    )
    return OmeZarrPlate(group_handler)


def create_empty_plate(
    store: StoreOrGroup,
    name: str,
    images: list[ImageInWellPath] | None = None,
    version: str = "0.4",
    cache: bool = False,
    overwrite: bool = False,
    parallel_safe: bool = True,
) -> OmeZarrPlate:
    """Initialize and create an empty OME-Zarr plate."""
    mode = "w" if overwrite else "w-"
    group_handler = ZarrGroupHandler(
        store=store, cache=True, mode=mode, parallel_safe=False
    )
    meta_handler = get_plate_meta_handler(group_handler, version=version)
    plate_meta = NgioPlateMeta.default_init(
        name=name,
        version=version,
    )
    meta_handler.write_meta(plate_meta)

    if images is not None:
        plate = OmeZarrPlate(group_handler)
        for image in images:
            plate.add_image(
                row=image.row,
                column=image.column,
                image_path=image.path,
                acquisition_id=image.acquisition_id,
                acquisition_name=image.acquisition_name,
            )
    return open_omezarr_plate(
        store=store,
        cache=cache,
        mode="r+",
        parallel_safe=parallel_safe,
    )



from collections.abc import Collection

import pandas as pd
from pandas import DataFrame

from ngio.tables.backends._abstract_backend import AbstractTableBackend
from ngio.utils import NgioFileNotFoundError


class JsonTableBackend(AbstractTableBackend):
    """A class to load and write small tables in the zarr group .attrs (json)."""

    @staticmethod
    def backend_name() -> str:
        """The name of the backend."""
        return "experimental_json_v1"

    @staticmethod
    def implements_anndata() -> bool:
        """Whether the handler implements the anndata protocol."""
        return False

    @staticmethod
    def implements_dataframe() -> bool:
        """Whether the handler implements the dataframe protocol."""
        return True

    def load_columns(self) -> list[str]:
        """List all labels in the group."""
        return list(self.load_as_dataframe().columns)

    def _get_table_group(self):
        try:
            table_group = self._group_handler.get_group(path="table")
        except NgioFileNotFoundError:
            table_group = self._group_handler.group.create_group("table")
        return table_group

    def load_as_dataframe(self, columns: Collection[str] | None = None) -> DataFrame:
        """List all labels in the group."""
        table_group = self._get_table_group()
        table_dict = dict(table_group.attrs)
        data_frame = pd.DataFrame.from_dict(table_dict)
        if columns is not None:
            data_frame = data_frame[columns]
        return data_frame

    def write_from_dataframe(
        self, table: DataFrame, metadata: dict | None = None
    ) -> None:
        """Consolidate the metadata in the store."""
        table_group = self._get_table_group()
        table_group.attrs.clear()
        table_group.attrs.update(table.to_dict())
        if metadata is not None:
            self._group_handler.write_attrs(metadata)



from collections.abc import Collection
from pathlib import Path

from anndata import AnnData
from pandas import DataFrame

from ngio.tables.backends._abstract_backend import AbstractTableBackend
from ngio.tables.backends._anndata_utils import (
    anndata_to_dataframe,
    custom_read_zarr,
    dataframe_to_anndata,
)
from ngio.utils import NgioValueError


class AnnDataBackend(AbstractTableBackend):
    """A class to load and write tables from/to an AnnData object."""

    @staticmethod
    def backend_name() -> str:
        """The name of the backend."""
        return "anndata_v1"

    @staticmethod
    def implements_anndata() -> bool:
        """Whether the handler implements the anndata protocol."""
        return True

    @staticmethod
    def implements_dataframe() -> bool:
        """Whether the handler implements the dataframe protocol."""
        return True

    def load_columns(self) -> list[str]:
        """List all labels in the group."""
        return list(self.load_as_dataframe().columns)

    def load_as_anndata(self, columns: Collection[str] | None = None) -> AnnData:
        """Load the metadata in the store."""
        anndata = custom_read_zarr(self._group_handler._group)
        if columns is not None:
            raise NotImplementedError(
                "Selecting columns is not implemented for AnnData."
            )
        return anndata

    def load_as_dataframe(self, columns: Collection[str] | None = None) -> DataFrame:
        """List all labels in the group."""
        dataframe = anndata_to_dataframe(
            self.load_as_anndata(),
            index_key=self._index_key,
            index_type=self._index_type,
        )
        if columns is not None:
            dataframe = dataframe[columns]
        return dataframe

    def write_from_dataframe(
        self, table: DataFrame, metadata: dict | None = None
    ) -> None:
        """Consolidate the metadata in the store."""
        anndata = dataframe_to_anndata(table, index_key=self._index_key)
        self.write_from_anndata(anndata, metadata)

    def write_from_anndata(self, table: AnnData, metadata: dict | None = None) -> None:
        """Consolidate the metadata in the store."""
        store = self._group_handler.store
        if not isinstance(store, str | Path):
            raise NgioValueError(
                "To write an AnnData object the store must be a local path/str."
            )

        store = Path(store) / self._group_handler.group.path
        table.write_zarr(store)
        if metadata is not None:
            self._group_handler.write_attrs(metadata)



"""Next Generation file format IO."""

from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version("ngio")
except PackageNotFoundError:  # pragma: no cover
    __version__ = "uninstalled"
__author__ = "Lorenzo Cerrone"
__email__ = "lorenzo.cerrone@uzh.ch"

from ngio.common import ArrayLike, Dimensions
from ngio.hcs import OmeZarrPlate, create_empty_plate, open_omezarr_plate
from ngio.images import (
    Image,
    Label,
    OmeZarrContainer,
    create_empty_omezarr,
    create_omezarr_from_array,
    open_image,
    open_omezarr_container,
)
from ngio.ome_zarr_meta.ngio_specs import AxesSetup, ImageInWellPath, PixelSize

__all__ = [
    "ArrayLike",
    "AxesSetup",
    "Dimensions",
    "Image",
    "ImageInWellPath",
    "Label",
    "OmeZarrContainer",
    "OmeZarrPlate",
    "PixelSize",
    "create_empty_omezarr",
    "create_empty_plate",
    "create_omezarr_from_array",
    "open_image",
    "open_omezarr_container",
    "open_omezarr_plate",
]
